from typing import List

from prometheus_client import (
    Counter,
    Histogram,
    Summary,
    Gauge,
    Info,
    Enum
)

from prometheusrock.middleware import MetricsStorage
from prometheusrock.singleton import SingletonMeta


class Metric:
    def __init__(self, metric: Counter, function: object, metric_type: str, spent_time: float = 0):
        """
        Storage of metric properties

        Args:
            metric (Info): one of prometheusrock metric types.
                Info type was chosen because we cant import from `prometheus_client' base class.
            function (object): function, that returns data that you want to add. It MUST return list or tuple.
                It may return dict if all keys are the same with metric.
            metric_type (str): assigned metric type
            spent_time (float): amount of time that was spent on request

        """
        self.metric = metric
        self.function = function
        self.metric_type = metric_type
        self.spent_time = spent_time


class AddMetric:
    def __init__(self,
                 function: object,
                 metric_name: str = 'user_metric',
                 metric_description: str = 'description of user metric',
                 labels: List[str] = ["info"],
                 metric_type: str = ''):
        """
        Add your custom metric for Prometheus. Constructor for dynamic class

        Args:
            function (object): function, that returns data that you want to add. It MUST return list or tuple.
                It may return dict if all keys are the same with metric.
            metric_name (str): unique metric name, must be ONE-WORDED (e.g. unique_metric_name)
            metric_description (str): description of your metric
            labels (List[str]): list of labels for your metric
            metric_type (str): metric that you want to use: counter, histogram, summary, info, enum, gauge.
                More info about metric types you can find [here](https://github.com/prometheus/client_python)

        """
        params = self._ParamStorage(metric_name=metric_name,
                                    metric_description=metric_description,
                                    function=function,
                                    metric_type=metric_type,
                                    labels=labels)

        self._GetTheFlame(params)

    class _ParamStorage:
        def __init__(self, **kwargs):
            self.metric_name = kwargs.get('metric_name', '')
            self.metric_description = kwargs.get('metric_description', '')
            self.function = kwargs.get('function', '')
            self.labels = kwargs.get('labels', '')
            self.metric_type = kwargs.get('metric_type', '').lower()

    class _GetTheFlame:
        def __init__(self, params: '_ParamStorage'):
            self.params = params
            _custom_metric_builder = self.generate_metric_class()
            _custom_metric_builder(self.params)

        def generate_metric_class(self):
            return SingletonMeta(
                self.params.metric_name,
                (),
                {
                    "__init__": self._metric_constructor
                }
            )

        def _metric_constructor(self, params: '_ParamStorage'):
            types = {
                'counter': Counter,
                'histogram': Histogram,
                'summary': Summary,
                'info': Info,
                'enum': Enum,
                'gauge': Gauge
            }

            metric_pool = MetricsStorage()

            if types.get(params.metric_type.lower()):
                try:
                    metric = types[params.metric_type.lower()](
                        params.metric_name,
                        params.metric_description,
                        params.labels
                    )
                except ValueError:
                    raise ValueError(f"You already registered metric with name {params.metric_name}!")
            else:
                raise TypeError("""Invalid metric type! Choose of the following metric types:
                        1. counter
                        2. histogram
                        3. gauge
                        4. summary
                        5. info
                        6. enum
                        """)

            metric_pool.custom_metrics.append(
                Metric(
                    metric=metric,
                    function=params.function,
                    metric_type=params.metric_type.lower()
                )
            )



import os

from prometheus_client import (
    generate_latest,
    CONTENT_TYPE_LATEST,
    REGISTRY,
    multiprocess,
    CollectorRegistry
)
from starlette import status
from starlette.requests import Request
from starlette.responses import Response


def metrics_route(request: Request):
    """
    Endpoint for Prometheus metrics_route. Code taken from prometheus_client examples.

    Examples:
        app.add_middleware(PrometheusMiddleware, {params})

        app.add_route("/metrics_route", metrics_route)

    """
    registry = REGISTRY
    if 'prometheus_multiproc_dir' in os.environ:
        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)

    data = generate_latest(registry)
    response_headers = {
        'Content-type': CONTENT_TYPE_LATEST,
        'Content-Length': str(len(data))
    }
    return Response(data, status_code=status.HTTP_200_OK, headers=response_headers)



import time
import inspect
from typing import List, Tuple

from prometheus_client import (
    Counter,
    Histogram
)
from starlette import status
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.types import ASGIApp

from prometheusrock.singleton import SingletonMeta


class MetricsStorage(metaclass=SingletonMeta):
    def __init__(self, labels: List[str] = ["method", "path", "status_code", "headers", "app_name"]):
        self.labels = labels
        self.REQUEST_COUNT = Counter(
            "requests_total",
            "Total HTTP requests",
            labels,
        )

        self.REQUEST_TIME = Histogram(
            "request_processing_time",
            "HTTP request processing time in seconds",
            labels,
        )

        self.custom_metrics = []


class PrometheusMiddleware(BaseHTTPMiddleware):
    def __init__(self,
                 app: ASGIApp,
                 app_name: str = "ASGIApp",
                 additional_headers: List[str] = [],
                 remove_labels: List[str] = [],
                 skip_paths: List[str] = ['/metrics']):

        """
        Configuration class for PrometheusPilgrimage middleware.

        Args:
            app (ASGIApp): instance of App (starlette/FastAPI)
            app_name (str): desirable name. default = "ASGIApp"
            additional_headers (List[str]): headers that you want to watch. Default ones - ["user-agent", "host"]
            remove_labels (Tuple[str]): labels that you want to remove.
                Default labels - ["method", "path", "status_code", "headers", "app_name"]
            skip_paths (List[str]): if you dont want to log events on specific paths, pass them here. Default on '/metrics_route'

        """
        if not isinstance(additional_headers, list):
            raise TypeError("additional_headers must be list!")
        if not isinstance(remove_labels, list):
            raise TypeError("remove_labels must be list!")
        if not isinstance(skip_paths, list):
            raise TypeError("skip_paths must be list!")

        super().__init__(app)

        base_labels = ["method", "path", "status_code", "headers", "app_name"]
        [base_labels.remove(item) for item in remove_labels if item in base_labels]
        labels = list(set([item.lower() for item in base_labels]))
        if len(labels) == 0:
            raise ValueError("Labels cant be empty!")

        self.metrics = MetricsStorage(labels)

        self.app_name = app_name

        base_headers = ["user-agent", "host"]
        needed_headers = base_headers + additional_headers
        self.needed_headers = list(set([item.lower() for item in needed_headers]))

        self.skip_paths = skip_paths

    async def dispatch(self, request, call_next):
        if request.url.path not in self.skip_paths:
            method = request.method
            path = request.url.path
            headers = {key.lower(): value for key, value in request.headers.items() if
                       key.lower() in self.needed_headers}
            begin = time.time()

            status_code = status.HTTP_408_REQUEST_TIMEOUT

            try:
                response = await call_next(request)
                status_code = response.status_code

            except Exception as e:
                raise e

            finally:
                spent_time = time.time() - begin

                labels = {
                    "method": method,
                    "path": path,
                    "status_code": status_code,
                    "headers": headers,
                    "app_name": self.app_name
                }

                final_labels = {
                    item: labels.get(item) for item in labels.keys() if
                    item in self.metrics.labels
                }
                self.metrics.REQUEST_COUNT.labels(**final_labels).inc()
                self.metrics.REQUEST_TIME.labels(**final_labels).observe(spent_time)
                for metric_key in self.metrics.custom_metrics:
                    metric_key.spent_time = spent_time
                    if inspect.iscoroutinefunction(metric_key.function):
                        await metric_key.function(metric_key)
                    else:
                        metric_key.function(metric_key)

            return response
        else:
            try:
                return await call_next(request)
            except Exception as e:
                raise e



from setuptools import setup

setup(
    name='prometheusrock',
    version='0.1.0',
    author='Nikita Kozhushko (kozhushman)',
    author_email='kozhushman@gmail.com',
    packages=['prometheusrock'],
    license='MIT License',
    url="https://github.com/kozhushman/prometheusrock",
    description='Prometheus middleware for Starlette and FastAPI',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "starlette",
        "prometheus_client"
    ],
)
